/*  Copyright (C) 2008 Niels Richard Hansen. 

    This file is part of StemSearch.

    StemSearch is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    StemSearch is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "StructSearch.h"
#include <cmath>
#include <algorithm>
#include <sstream>

/* Constructors, destructors etc. */

Island::Island()
{
	Pos.resize(2);
	Anchor.resize(2);
	Score =0;
	Evalue =0;
	GenomeName="";
	BandSize=0;
	Sequence="";
	Structure3="";
	Structure5="";
	IslandNr=0;
}

StructSearch::StructSearch(long band_size, double t, const char* parameter_filename, bool Lmethod, bool w, long s, long e)
{
	eos = false;
	band = band_size;
	hplooppen = Lmethod;
	warnings=w;
	G.warn(w);
	read_par(parameter_filename);
	threshold = max((t+logK)/lambda, (double) 0);
	HPLoopPen.resize(band);
	start = max(s, (long) 0) + 4;
	end = e;
	store_random_seq = true;
}

/*
The four central dynamic programming recursions:                              
                                                                              
computeST   A straight forward computation of the two dynamic programming     
            score matrices (SS, LS) including traceback information (ST, LT)  
            and position of the maximum (Mposi, Mposj). The actual traceback  
            from the position of maximal score is carried out separately.     
                                                                              
computeSIT  As computeST but with island declumping (SI, LI). Multiple        
            suboptimal positions are stored.                                   
                                                                              
computeEST  As computeST but with a memory efficient implementation that does 
            not store the entire score matrices. The actual traceback is  
            carried out "on the fly".                                   
                                                                              
computeESIT Combines computeSIT and computeEST so that island declumping is   
            done using a memory efficient implementation. This is the default 
            and in cases of intended use the relevant algorithm.         
*/

void StructSearch::computeST()
{
  long i,j,i1,j1,cs,cl,m,ip,jm;
	double ts[3],tl[5];
	long MPosi=0,MPosj=0;
	
	SIT = 1;
	seqstr.resize(1);
	SS.resize(l+1,l+1,L);
	ST.resize(l+1,l+1,L);
	LS.resize(l+1,l+1,L);
	LT.resize(l+1,l+1,L);

	j = start-1;
	
	SS(j-1,j) = HPLoopPen[2];
	LS(j-1,j) = HPLoopPen[2];
	SS(j-2,j) = HPLoopPen[3];
	LS(j-2,j) = HPLoopPen[3];
		
	for(j=start;j<=l;j++)
	{	
		jm = j-1;
		
		SS(jm,j) = HPLoopPen[2];
		LS(jm,j) = HPLoopPen[2];
		SS(j-2,j) = HPLoopPen[3];
		LS(j-2,j) = HPLoopPen[3];
			
		for(i=j-3;i>max(j-L,(long) 0);i--)
		{
			ip = i+1;
						
			cl=0;
			cs=0;
			
			ts[0] = HPLoopPen[j-i+1];
			tl[0] = HPLoopPen[j-i+1];
			
			//Computing scores for the two to-stack and four to-loop transitions 
			
			if(G[i] != 4 && G[ip] != 4 && G[j] != 4 && G[jm] != 4)     //Check if any of the involved letters are non-standard 
			{
				i1 = G[ip] + 4*G[jm];
				j1 = G[i] + 4*G[j];
				
				if(StackMask(i1,j1))                                 //Check if the stack is allowed
				{
					ts[1] = SS(ip,jm) + SScore(i1,j1);
					ts[2] = LS(ip,jm) + SScore(i1,j1);
				}
				else 
				{
					ts[1]=ts[0]; 
					ts[2]=ts[0];	
				}
				 
				tl[1] = SS(i,jm) + Li; 
				tl[2] = SS(ip,j) + Li; 
				tl[3] = LS(i,jm) + Le; 
				tl[4] = LS(ip,j) + Le; 
				
			       
				
				//Selecting the highest scoring transitions -- to-loop and to-stack resp.
				
				if(tl[cl] < tl[1]) cl=1;
				if(tl[cl] < tl[2]) cl=2;
				if(tl[cl] < tl[3]) cl=3;
				if(tl[cl] < tl[4]) cl=4;
				
				if(ts[cs] < ts[1]) cs=1;
				if(ts[cs] < ts[2]) cs=2;
			}
			else
			{
				cl = 0;
				cs = 0;
			}
			
				
			LS(i,j) = tl[cl];
			LT(i,j) = cl;
			
			SS(i,j) = ts[cs];
			ST(i,j) = cs;
			
			if(ts[cs] > SS(MPosi,MPosj))
			{
				seqstr[0].pos(i,j);
				MPosi=i;
				MPosj=j;
			}
			
		}
		
	}
}


void StructSearch::computeSIT()
{
  long i,j,i1,j1,cs,cl,m,ip,jm,imL,jmL,MI=0;
	double ts[5],tl[5];
	
	SIT = 2;
	seqstr.resize(0);
	SS.resize(l+1,l+1,L);
	ST.resize(l+1,l+1,L);
	SI.resize(l+1,l+1,L);
	LS.resize(l+1,l+1,L);
	LT.resize(l+1,l+1,L);
	LI.resize(l+1,l+1,L);
	MaxPosi.clear();
	MaxPosj.clear();
	MaxPosi.resize(2*(l+1)*L);
	MaxPosj.resize(2*(l+1)*L);

	j = start-1;
		
		MI++;
		SI(j-1,j) = MI;
		LI(j-1,j) = MI;
		SS(j-1,j) = HPLoopPen[2];
		LS(j-1,j) = HPLoopPen[2];
		MI++;
		SI(j-2,j) = MI;
		LI(j-2,j) = MI;
		SS(j-2,j) = HPLoopPen[3];
		LS(j-2,j) = HPLoopPen[3];
	
	for(j=start;j<=l;j++)
	{
	  	jm = j-1;
		
		MI++;
		SI(jm,j) = MI;
		LI(jm,j) = MI;
		SS(jm,j) = HPLoopPen[2];
		LS(jm,j) = HPLoopPen[2];
		MI++;
		SI(j-2,j) = MI;
		LI(j-2,j) = MI;
		SS(j-2,j) = HPLoopPen[3];
		LS(j-2,j) = HPLoopPen[3];
			
		for(i=j-3;i>max(j-L,(long) start-4);i--)
		  {

			ip = i+1;
			
			cl=0;
			cs=0;
			
			ts[0] = HPLoopPen[j-i+1];
			tl[0] = HPLoopPen[j-i+1];
			
			//Computing scores for the two to-stack and four to-loop transitions 
			
			if(G[i] != 4 && G[ip] != 4 && G[j] != 4 && G[jm] != 4)     //Check if any of the involved letters are non-standard 
			{
				i1 = G[ip] + 4*G[jm];
				j1 = G[i] + 4*G[j];
				
				if(StackMask(i1,j1))                                 //Check if the stack is allowed
				{
					ts[1] = SS(ip,jm) + SScore(i1,j1);
					ts[2] = LS(ip,jm) + SScore(i1,j1);
				}
				else 
				{
					ts[1]=ts[0]; 
					ts[2]=ts[0];	
				}
				 
				tl[1] = SS(i,jm) + Li; 
				tl[2] = SS(ip,j) + Li; 
				tl[3] = LS(i,jm) + Le; 
				tl[4] = LS(ip,j) + Le; 
				
			       
				
				//Selecting the highest scoring transitions -- to-loop and to-stack resp.
				
				if(tl[cl] < tl[1]) cl=1;
				if(tl[cl] < tl[2]) cl=2;
				if(tl[cl] < tl[3]) cl=3;
				if(tl[cl] < tl[4]) cl=4;
				
				if(ts[cs] < ts[1]) cs=1;
				if(ts[cs] < ts[2]) cs=2;
			}
			else
			{
				cl = 0;
				cs = 0;
			}
						
				
			LS(i,j) = tl[cl];
			LT(i,j) = cl;
			
			SS(i,j) = ts[cs];
			ST(i,j) = cs;

			
			//Filling the island matrices and checking for exceedances of threshold
			
		 	switch (cs)
			  {
			  case 0:                               //A new island is generated
			    {
				MI++;
				SI(i,j) = MI;			//A labeling convention for islands
				MaxPosi[SI(i,j)] = i;				
				MaxPosj[SI(i,j)] = j;
				break;
			    }
			  case 1:
			    {
			      SI(i,j) = SI(ip,jm);
			      break;
			    }
			  case 2:
			    {
			      SI(i,j) = LI(ip,jm);
			      break;
			    }
			    break;
			  }
			
			if((ts[cs] > SS(MaxPosi[SI(i,j)],MaxPosj[SI(i,j)])) && (ts[cs] > threshold))
			      {	
				MaxPosi[SI(i,j)] = i; 
				MaxPosj[SI(i,j)] = j;
			      }
			 
			    
			  
   
			switch(cl)
			  {
			    case 0:                               //A new island is generated
			    {
				MI++;
				LI(i,j) = MI;			//A labeling convention for islands
				MaxPosi[LI(i,j)] = i;				
				MaxPosj[LI(i,j)] = j;
				break;
			    }
			  case 1:
			    {
			      LI(i,j) = SI(i,jm);
			      break;
			    }
			  case 2:
			    {
			      LI(i,j) = SI(ip,j);
			      break;
			    }
			  case 3:
			    {
			      LI(i,j) = LI(i,jm);
			      break;
			    }
			  case 4:
			    {
			      LI(i,j) = LI(ip,j);
			    break;
			    }
			    break;
			  }

//                      imL = i-2*L;
// 			jmL = j-2*L;

			
// 		 	if((i>2*L) && (SS(MaxPosi[SI(imL,jmL)],MaxPosj[SI(imL,jmL)]) > threshold))
//  			{

//  				seqstr.resize(seqstr.size()+1);
//  				seqstr.back().pos(MaxPosi[SI(imL,jmL)],MaxPosj[SI(imL,jmL)]); 
//  				MaxPosi[SI(imL,jmL)]=0;
//  				MaxPosj[SI(imL,jmL)]=0;
//  			}
			
			
 		}
 	}

	for(j=start;j<=l;j++)
	{
		for(i=j-3;i>max(j-L,(long) start-4);i--)
		{

			if(SS(MaxPosi[SI(i,j)],MaxPosj[SI(i,j)]) > threshold)
			{
				seqstr.resize(seqstr.size()+1);
				seqstr.back().pos(MaxPosi[SI(i,j)],MaxPosj[SI(i,j)]); 
				MaxPosi[SI(i,j)]=0;
				MaxPosj[SI(i,j)]=0;
			}
		}
	}
	
}
	
void StructSearch::computeEST()
{
  long i,j,ii,jj,i1,j1,cl,cs,m,ip,ipp,jm,jmL;
	double ts[3],tl[5],MaxScore=0;
	long MPosj= 0; 
	long LL = 2*L +1;

	SIT = 3;
	seqstr.resize(1);
	ESS.init(L,LL);
	ELS.init(L,LL);
	EST.init(L,LL);
	ELT.init(L,LL);

	jj = (start-1) % (LL);
	
	ESS(L-1,jj) = HPLoopPen[2];
	ELS(L-1,jj) = HPLoopPen[2];
	ESS(L-2,jj) = HPLoopPen[3];
	ELS(L-2,jj) = HPLoopPen[3];
	
	
	for(j=start;j<=l;j++)
	{
		jm = (j-1) % (LL);
		jj = j % (LL);

		ESS(L-1,jj) = HPLoopPen[2];
		ELS(L-1,jj) = HPLoopPen[2];
		ESS(L-2,jj) = HPLoopPen[3];
		ELS(L-2,jj) = HPLoopPen[3];
			
		for(i=j-3;i>max(j-L,(long) 0);i--)
		{
			ii = i+L-j;
			ip = ii+1;
			ipp = ii+2;

			
			cl=0;
			cs=0;
			
			ts[0] = HPLoopPen[j-i+1];
			tl[0] = HPLoopPen[j-i+1];
			
			//Computing scores for the two to-stack and four to-loop transitions 
			
			if(G[i] != 4 && G[i+1] != 4 && G[j] != 4 && G[j-1] != 4)     //Check if any of the involved letters equal N 
			{
				i1 = G[i+1] + 4*G[j-1];
				j1 = G[i] + 4*G[j];
				
					
				if(StackMask(i1,j1))                                 //Check if the stack is allowed
				{
					ts[1] = ESS(ipp,jm) + SScore(i1,j1);
					ts[2] = ELS(ipp,jm) + SScore(i1,j1);
				}
				else 
				{
					ts[1]=ts[0]; 
					ts[2]=ts[0];	
				}
				 
				tl[1] = ESS(ip,jm) + Li; 
				tl[2] = ESS(ip,jj) + Li; 
				tl[3] = ELS(ip,jm) + Le; 
				tl[4] = ELS(ip,jj) + Le; 
				
			       
				
				//Selecting the highest scoring transitions -- to-loop and to-stack resp.
				
				if(tl[cl] < tl[1]) cl=1;
				if(tl[cl] < tl[2]) cl=2;
				if(tl[cl] < tl[3]) cl=3;
				if(tl[cl] < tl[4]) cl=4;
				
				if(ts[cs] < ts[1]) cs=1;
				if(ts[cs] < ts[2]) cs=2;
			}
			else
			{
				cl = 0;
				cs = 0;
			}
			
				
			ELS(ii,jj) = tl[cl];
			ELT(ii,jj) = cl;
			
			ESS(ii,jj) = ts[cs];
			EST(ii,jj) = cs;
			
			if(ts[cs] > MaxScore) 
			{	
				seqstr.back().pos(i,j);
				MPosj = j; 
				MaxScore = ts[cs];
			}
		}
		

		if(MPosj == j-L-1) Etraceback(seqstr[0]);
	}
	if(MPosj > j-L-1) Etraceback(seqstr[0]);
}


void StructSearch::computeESIT()
{
  long i,j,ii,jj,i1,j1,cl,cs,m,ip,ipp,jm,jmL;
	double ts[3],tl[5];
	vector<long> MPosi(2*L*(L+1)), MPosj(2*L*(L+1));


	long LL = 2*L+1;
	SIT = 4;
	seqstr.resize(0);
	ESS.init(L+1,LL);
	EST.init(L,LL);
	ESI.init(L,LL);
	ELS.init(L,LL);
	ELT.init(L,LL);
	ELI.init(L,LL);

	jj = (start-1) % (LL);
	

	ESS(L-1,jj) = HPLoopPen[2];
	ELS(L-1,jj) = HPLoopPen[2];
	ESI(L-1,jj) = L + jj*L;
	ELI(L-1,jj) = ESI(L-1,jj);
	ESS(L-2,jj) = HPLoopPen[3];
	ELS(L-2,jj) = HPLoopPen[3];
	ESI(L-2,jj) = L-1 + jj*L;
	ELI(L-2,jj) = ESI(L-2,jj);
	
	for(j=start;j<=l;j++)
	{
		jm = (j-1) % (LL);
		jj = j % (LL);
		jmL = (j+L)%(LL);


		ESS(L-1,jj) = HPLoopPen[2];
		ELS(L-1,jj) = HPLoopPen[2];
		ESI(L-1,jj) = L + jj*L;
		ELI(L-1,jj) = ESI(L-1,jj);
		ESS(L-2,jj) = HPLoopPen[3];		
		ELS(L-2,jj) = HPLoopPen[3];
		ESI(L-2,jj) = L-1 + jj*L;
		ELI(L-2,jj) = ESI(L-2,jj);

		for(i=j-3;i>max(j-L,(long) start-4);i--)
		{
			ii = i+L-j;
			ip = ii+1;
			ipp = ii+2;
					      
			cl=0;
			cs=0;
			
			ts[0] = HPLoopPen[j-i+1];
			tl[0] = HPLoopPen[j-i+1];
			
			//Computing scores for the two to-stack and four to-loop transitions 
			
			if(G[i] != 4 && G[i+1] != 4 && G[j] != 4 && G[j-1] != 4)     //Check if any of the involved letters equal N 
			{
				i1 = G[i+1] + 4*G[j-1];
				j1 = G[i] + 4*G[j];
				
					
				if(StackMask(i1,j1))                                 //Check if the stack is allowed
				{
					ts[1] = ESS(ipp,jm) + SScore(i1,j1);
					ts[2] = ELS(ipp,jm) + SScore(i1,j1);
				}
				else 
				{
					ts[1]=ts[0]; 
					ts[2]=ts[0];	
				}
				 
				tl[1] = ESS(ip,jm) + Li; 
				tl[2] = ESS(ip,jj) + Li; 
				tl[3] = ELS(ip,jm) + Le; 
				tl[4] = ELS(ip,jj) + Le; 
				
			       
				
				//Selecting the highest scoring transitions -- to-loop and to-stack resp.
				
				if(tl[cl] < tl[1]) cl=1;
				if(tl[cl] < tl[2]) cl=2;
				if(tl[cl] < tl[3]) cl=3;
				if(tl[cl] < tl[4]) cl=4;
				
				if(ts[cs] < ts[1]) cs=1;
				if(ts[cs] < ts[2]) cs=2;
			}
			else
			{
				cl = 0;
				cs = 0;
			}
			
				
			ELS(ii,jj) = tl[cl];
			ELT(ii,jj) = cl;
			
			ESS(ii,jj) = ts[cs];
			EST(ii,jj) = cs;
			
			//Filling the island matrices and checking for exceedances of threshold

			

	
			switch (cs)
			  {
			     case 0:                               //A new island is generated
			    {
			        ESI(ii,jj) = ii + L*jj;		//A labeling convention for islands
				MPosi[ESI(ii,jj)] = i;				
				MPosj[ESI(ii,jj)] = j;
				break;
				}
			  case 1:
			    {
			      ESI(ii,jj) = ESI(ipp,jm);
			      break;
			    }
			  case 2:
			    {
			      ESI(ii,jj) = ELI(ipp,jm);
			      break;
			    }
			    break;
			  }
			
			if((ts[cs] > threshold) && (ts[cs] > ESS(MPosi[ESI(ii,jj)]+L-MPosj[ESI(ii,jj)],MPosj[ESI(ii,jj)]%(LL))))
			  {
			    MPosi[ESI(ii,jj)] = i; 
			    MPosj[ESI(ii,jj)] = j;
			  }
			
			switch(cl)
			  {
			     case 0:                               //A new island is generated
			    {
			        ELI(ii,jj) = ii + L*jj;		//A labeling convention for islands
				MPosi[ELI(ii,jj)] = i;				
				MPosj[ELI(ii,jj)] = j;
				break;
			    }
			  case 1:
			    {
			      ELI(ii,jj) = ESI(ip,jm);
			      break;
			    }
			  case 2:
			    {
			      ELI(ii,jj) = ESI(ip,jj);
			      break;
			    }
			  case 3:
			    {
			      ELI(ii,jj) = ELI(ip,jm);
			      break;
			    }
			  case 4:
			    {
			      ELI(ii,jj) = ELI(ip,jj);
			    break;
			    }
			    break;
			  }

				
		        if(i > L+1)
			  {
			    i1 = MPosi[ESI(ii,jmL)]+L-MPosj[ESI(ii,jmL)];
			    j1 = MPosj[ESI(ii,jmL)]%(LL);

			    if(ESS(i1,j1) > threshold) 
			      {
					seqstr.resize(seqstr.size()+1);
					seqstr.back().pos(MPosi[ESI(ii,jmL)],MPosj[ESI(ii,jmL)]);
					Etraceback(seqstr.back());
 					ESS(i1,j1) = 0;
				}
			}			

			
		}
	}
	
	
	for(j=l-L+1;j<=l;j++)
	{
		jmL = j % (LL);
		
		for(i=j-3;i>max(j-L,(long) 0);i--)
		{
			ii = i+L-j;
			i1 = MPosi[ESI(ii,jmL)]+L-MPosj[ESI(ii,jmL)];
			j1 = MPosj[ESI(ii,jmL)]%(LL);
			
			if(ESS(i1,j1) > threshold) 
			{
				seqstr.resize(seqstr.size()+1);
				seqstr.back().pos(MPosi[ESI(ii,jmL)],MPosj[ESI(ii,jmL)]);
				Etraceback(seqstr.back());
 				ESS(i1,j1) = 0;
			}
		}
	}
}			
	

/* 
Traceback algorithms:                                                         
Etraceback is for traceback "on the fly" for the memory efficient algorithm   
traceback is for traceback using the full, stored score matrices              
*/
	
void StructSearch::Etraceback(Island& island)
{
	long i=island[0],j=island[1],ii=i+L-j,jj=j%(2*L+1);
	int backtrack=1; /*
			   backtrack = 0: terminate traceback, 
			   backtrack = 1: use stack traceback matrix, 
			   backtrack=2: use loop traceback matrix */

	double score = lambda*ESS(ii,jj) - logK;
	string Seqtemp,Str5temp,Str3temp,loop;
	stringstream tmp;
	
	island.genome() = G.name();
	island.bandsize() = L;
	island.score() = score;
	if(hplooppen)
	  {
	    island.evalue() = exp( log((double) l-start+4) - score);
	  }
	else
	  {
	    island.evalue() = exp( log((double) (l-start+4 - L/2)) + log((double) (L-1)) - score);
	  }
		
	while(backtrack > 0)
	{
		if(jj == -1) jj = 2*L;
		switch (backtrack)
		{
		case 1:
		  switch (EST(ii,jj))
		    {
		    case 1: 
		      Str5temp += "(";
		      Str3temp = ")" + Str3temp;
		      jj--;
		      ii = ii+2;
		      backtrack = 1;
		      break;
		    case 2:
		      Str5temp += "(";
		      Str3temp = ")" + Str3temp;
		      jj--;
		      ii = ii+2;
		      backtrack = 2;
		      break;
		    default:
		      backtrack=0;
		      break;	
		    }
		break;
		case 2:
			switch (ELT(ii,jj))
			{
				case 1: 
					Str3temp = "." + Str3temp;
					jj--;
					ii++;
					backtrack = 1;
					break;
				case 2:
					Str5temp += ".";
					ii++;
					backtrack = 1;
					break;
				case 3: 
					Str3temp = "." + Str3temp;
					jj--;
					ii++;
					backtrack=2;
					break;
				case 4: 
					Str5temp += ".";
					ii++;
					backtrack=2;
					break;
				default:
					backtrack=0;
					break;	
			}
			break;
		default:
			backtrack=0;
			break;	
		}
	}
	
	i += Str5temp.size();
	j -= Str3temp.size();
	
	island.anchor(i,j);	
	island.sequence() = G.RNAsegment(island[0],island[1]);
	island.structure5() = Str5temp; 
	island.structure3() = Str3temp;
}

void StructSearch::traceback(Island& island)
{
	long i=island[0],j=island[1];
	int backtrack=1; /* 
			    backtrack = 0: terminate traceback, 
			    backtrack = 1: use stack traceback matrix, 
			    backtrack = 2: use loop traceback matrix */

	double score = lambda*SS(i,j) - logK;
	string Seqtemp,Str5temp,Str3temp,loop;
	stringstream tmp;
	
	island.genome() = G.name();
	island.bandsize() = L;
	island.islandnr() = SI(i,j);
	island.score() = score;
	if(hplooppen)
	  {
	    island.evalue() = exp( log((double) (l-start+4)) - score);
	  }
	else
	  {
	    island.evalue() = exp( log((double) (l-start+4 - L/2)) + log((double) (L-1)) - score);
	  }

	if(SIT==2) 
	{
		MaxPosi[SI(i,j)] = i;			//Restore the links that
		MaxPosj[SI(i,j)] = j;			//were removed by computeSIT
	}

	while(backtrack > 0)
	{
		switch (backtrack)
		{
		case 1:
			switch (ST(i,j))
			{
				case 1: 
					Str5temp += "(";
					Str3temp = ")" + Str3temp;
					j--;
					i++;
					backtrack = 1;
					break;
				case 2:
					Str5temp += "(";
					Str3temp = ")" + Str3temp;
					j--;
					i++;
					backtrack = 2;
					break;
				default:
					backtrack=0;
					break;	
			}
			break;
		case 2:
			switch (LT(i,j))
			{
				case 1: 
					Str3temp = "." + Str3temp;
					j--;
					backtrack = 1;
					break;
				case 2:
					Str5temp += ".";
					i++;
					backtrack = 1;
					break;
				case 3: 
					Str3temp = "." + Str3temp;
					j--;
					backtrack=2;
					break;
				case 4: 
					Str5temp += ".";
					i++;
					backtrack=2;
					break;
				default:
					backtrack=0;
					break;	
			}
			break;
		default:
			backtrack=0;
			break;	
		}
	}	
	
	island.anchor(i,j);
	
	
	island.sequence() = G.RNAsegment(island[0],island[1]);
	island.structure5() = Str5temp; 
	island.structure3() = Str3temp;
	
}

void StructSearch::traceback()
{
	if(SIT != 0)
	{
		switch (SIT)
		{
			case 1:
				traceback(seqstr[0]);
				break;
			case 2:
				for(long k=0;k<seqstr.size();k++)
					traceback(seqstr[k]);
				break;
		}
	sort(seqstr.begin(),seqstr.end(),str_order);
	}
	SIT = 0; 				//Traceback is not redone, if score matrices are not recomputed
}

/* Structure prediction using the score matrix and a greedy algorithm  */

void StructSearch::greedystr()
{	
	Island temp;
	
	long n,m,ii,jj,ai,aj,k,kk,i1,j1,i=seqstr[0][0],j=seqstr[0][1];
	double score=0;
	long s;
	string str(G.size(),'.');
	string tmpstr5,tmpstr3;
	
	for(k=0;k < seqstr.size();k++)
	{
		i1 = min(i,seqstr[k][0]);
		j1 = max(j,seqstr[k][1]);
		ai = seqstr[k][0];
		aj = seqstr[k](1);
		if(j1-i1 < L)
		{
			tmpstr5 = seqstr[k].structure5();
			n = tmpstr5.size(); 
			tmpstr3 = seqstr[k].structure3();
			m = tmpstr3.size();
			
			//Check for balance between anchor coordinates
			
			s=0;
 			for(kk=seqstr[k](0)-1;((kk<seqstr[k](1)) && (s >= 0));kk++)
 				s += (str[kk] == '(') - (str[kk] == ')');
			
			if(s==0)
			{
				//Find stem from anchor that fit into structure
				
				ii=n-1;
				jj=0;
				
				while((ii >=0) && (jj <m) && (str[ai-1+ii] == '.') && (str[aj+jj] == '.'))
				{
					if((tmpstr5[ii] == '(') && (tmpstr3[jj] == ')'))
					{
						ii--;
						jj++;
					}
					
					if((tmpstr5[ii] == '.') && (tmpstr3[jj] == ')'))
					{
						ii--;
					}
					
					if((tmpstr5[ii] == '(') && (tmpstr3[jj] == '.'))
					{
						jj++;
					}
					
					if((tmpstr5[ii] == '.') && (tmpstr3[jj] == '.'))
					{
						ii--;
						jj++;
					}
					
				}
				
				ii++;
				jj--;
				
				if((ii < n) && (jj >= 0))
				{
				
					while((tmpstr5[ii] == '.') || (tmpstr3[jj] == '.'))
					{
						if((tmpstr5[ii] == '.') && (tmpstr3[jj] == ')'))
						{
							ii++;
						}
						
						if((tmpstr5[ii] == '(') && (tmpstr3[jj] == '.'))
						{
							jj--;
						}
						
						if((tmpstr5[ii] == '.') && (tmpstr3[jj] == '.'))
						{
							ii++;
							jj--;
						}
					}
					
					if(SS(ai+ii,aj+1+jj)>threshold)
					{
						str.replace(ai+ii-1,n-ii,tmpstr5.substr(ii,n-ii));
						str.replace(aj,jj+1,tmpstr3.substr(0,jj+1));
						i = i1;
						j = j1;
						score += SS(ai+ii,aj+1+jj);
					}
				}
				
			}
			
		}
	}
	
	temp.sequence() = G.RNAsegment(1,G.size());
	temp.structure3() = str;
	temp.pos(i,j);
	temp.anchor(0,-1);
	temp.score() = score;
	temp.genome() = G.name() + " (greedy super structure) ";
	
	seqstr.push_front(temp);
}

vector<double> StructSearch::excesses(long b)
{
	long i,j;
	vector<double> tmp;
	
	for(long n=0;n<seqstr.size();n++)
	{
		i = seqstr[n](0);
		j = seqstr[n](1);
		if((i < G.size()-b) && (j >= b) && (j-i >= b))
			tmp.push_back(seqstr[n].score()-threshold);
	}
	
	return tmp;
}
	
/* Printing operations */

ostream& operator<<(ostream &s,Island &island)
{
	return s << island.pretty_print();
}

ostream& operator<<(ostream &s,StructSearch &StructObject)
{
	return s << StructObject.pretty_print();
}


string StructSearch::pretty_print(long nr,int output)
{
	stringstream SeqStr;
	long NR = nr;
	if((NR==0) || (NR > seqstr.size())) NR = seqstr.size();
	
	if(output==0)
	{
		for(long k=0;k<NR;k++)
			SeqStr << seqstr[k].pretty_print();
	}
	else if(output==1)
	{
		for(long k=0;k<NR;k++)
		  SeqStr << seqstr[k].pretty_print_minimal();
	}
	else if(output==2)
	{
		for(long k=0;k<NR;k++)
			SeqStr << seqstr[k].pretty_print_nl();
	}
	else if(output==3)
	  {
	    SeqStr << G.name() << "\t" << G.name() << "\t" << "contig\t" << 1 << "\t" << G.size() << "\t"  << ".\t.\t.\tContig " << G.name() << "\n";
		for(long k=0;k<NR;k++)
			SeqStr << seqstr[k].pretty_print_gff(k);
	}
	
	return SeqStr.str();
}

void StructSearch::print_score_matrices()
{
	cout.precision(3);
	SScore.print(5);
}	

/* Various Island implementations */

bool str_order(const Island &s1, const Island &s2)
{
	return (s1.score() > s2.score());
}



string Island::pretty_print()
{ 
	stringstream returnval;
	returnval.precision(4);

	returnval << ">" << GenomeName << "\nPosition: (" << Pos[0] << "," << Pos[1] << ") \t Anchor: (" << Anchor[0] << "," << Anchor[1] << ") \t Score: " << Score << " \t E-value: " << Evalue << "\n";	

		returnval << Sequence << "\n";
		returnval << Structure5 << string(Anchor[1]-Anchor[0]+1,'.') << Structure3 << "\n";

	return returnval.str();
}


string Island::pretty_print_minimal()
{ 
	stringstream returnval;
	returnval.precision(4);
	returnval << "\"" << GenomeName << "\""  << "\t" << Pos[0] << "\t" << Pos[1] << "\t"  << Anchor[0] << "\t" << Anchor[1] << "\t"  << Score << " \t" << Evalue << "\n";	
	return returnval.str();
}

string Island::pretty_print_nl()
{ 
	stringstream returnval;
	returnval.precision(4);
	returnval << ">" << GenomeName << "\nPosition: (" << Pos[0] << "," << Pos[1] << ") \t Anchor: (" << Anchor[0] << "," << Anchor[1] << ") \t Score: " << Score << " \t E-value: " << Evalue << "\n";	

	if(Anchor[1]-Anchor[0] < 15)
	{
		returnval << Sequence << "\n";
		returnval << Structure5 << string(Anchor[1]-Anchor[0]+1,'.') << Structure3 << "\n";
	}
	else 
	{	stringstream tmp;
		tmp << "[" << Anchor[1] - Anchor[0]+1 << "]";
		long n = Structure5.size()+2;
		long m = Structure3.size()+2;
		returnval <<  Sequence.substr(0,n) << tmp.str() << Sequence.substr(Sequence.size()-m,m) << "\n";
		returnval <<  Structure5 << ".." << tmp.str() << ".." << Structure3 << "\n";
	}	

	return returnval.str();
}


string Island::pretty_print_gff(long nr)
{ 
	stringstream returnval;
	returnval.precision(4);
	returnval << GenomeName << "\t" << "StemSearch" << VERSION << "\t" << "5'-stem\t" << Pos[0] << "\t" << Anchor[0]-1 << "\t" << Score << "\t+\t.\tStem_Loop_Rank" << nr+1 << "\n";
	returnval << GenomeName << "\t" << "StemSearch" << VERSION << "\t" << "3'-stem\t" << Anchor[1]+1 << "\t" << Pos[1] << "\t" << Score << "\t+\t.\tStem_Loop_Rank" << nr+1 << "\n";
	
	return returnval.str();
}

double& Island::score()
{ 
	return Score;
}

double Island::score() const
{ 
	return Score;
}

double& Island::evalue()
{ 
	return Evalue;
}

string& Island::genome()
{ 
	return GenomeName;
}

long& Island::bandsize()
{ 
	return BandSize;
}

long& Island::islandnr()
{ 
	return IslandNr;
}

string& Island::sequence()
{ 
	return Sequence;
}

string& Island::structure5()
{ 
	return Structure5;
}

string& Island::structure3()
{ 
	return Structure3;
}

void Island::anchor(long i,long j)
{ 
	Anchor[0] = i;
	Anchor[1] = j;
}

void Island::pos(long i,long j)
{ 
	Pos[0] = i;
	Pos[1] = j;
}

long Island::operator[](long i)
{
	return Pos[i];
}

long Island::operator()(long i)
{
	return Anchor[i];
}

/* Various StructSearch implementations  */

void StructSearch::error(char error_text[]) const
{
  cout << "\n" << error_text;
  exit(1);
}


void StructSearch::read(istream &st,char stopread)
{
	eos = G.read(st,stopread);
	if(end==0) l = G.size();
	else l = min(G.size(),end);
	if (band==0) L = G.size();
	else L = min(max(band,(long) 3),G.size());
}

void StructSearch::generate(long seed)
{
  if(end <= start) l = start;
  else l = end;
  G.Markov_generate(l,NullTP,seed);
  if (band==0) L = G.size();
  else L = min(max(band,(long) 3),G.size());

  eos = true;
  if(store_random_seq) 
    {
      ofstream file("random_seq");
      file << G.DNAsegment(1,l) << "\n";
    }
}

bool StructSearch::done()
{
	return eos;
}


void StructSearch::saveS()
{
	const char* tmp = (G.name() + ".sc").c_str();
	SS.eff_save(tmp);
}

void StructSearch::saveI()
{
	long i,j;
	long tmp = L-1;
	if(L%2 == 1) tmp = L;
	double temp;
				
	ofstream file((G.name() + ".is").c_str());
	if (!file) error("cannot open file");
	file.setf(ios_base::right,ios_base::adjustfield);
	
	for(j=tmp;j>=0;j=j-2)
	{
		for(i=0;i<j;i++)
			file << setw(5) << 0 << " ";
		for(i=0;i<SI.nr_rows()-j-1;i++)
		{
			temp = max(SS(MaxPosi[SI(i,i+j)],MaxPosj[SI(i,i+j)]),SS(MaxPosi[LI(i,i+j)],MaxPosj[LI(i,i+j)]));
			file << setw(5) << lambda*temp - logK*(temp >0) << " ";
			temp = max(SS(MaxPosi[SI(i,i+j+1)],MaxPosj[SI(i,i+j+1)]),SS(MaxPosi[LI(i,i+j+1)],MaxPosj[LI(i,i+j+1)]));
			file << setw(5) << lambda*temp - logK*(temp>0) << " ";
		}
		temp = max(SS(MaxPosi[SI(i,i+j)],MaxPosj[SI(i,i+j)]),SS(MaxPosi[LI(i,i+j)],MaxPosj[LI(i,i+j)]));
		file << setw(5) << lambda*temp - logK*(temp >0) << " ";
		for(i=0;i<j;i++)
			file << setw(5) << 0 << " ";
		file << "\n";
	}
	file << "\n";	
	
	
	ofstream file2((G.name() + ".r").c_str());
	if (!file2) error("cannot open file");
	
	double width = 10;
	double height=  max(min(5*min(band,G.size())/G.size(),(long) width),(long) 4);
	
	file2 << "IM <- read.table(\"" << G.name() << ".is\")\n";
	file2 << "IM2 <- t(IM[dim(IM)[1]:1,1:dim(IM)[2]])\n";
	file2 << "X11(width=" << width << ",height=" <<  height  << ")\n";
	file2 << "image(1:dim(IM2)[1],1:dim(IM2)[2],as.matrix(IM2),ann=F,col=gray(100:10/110))\n";
	file2 << "title(\"" << G.name() << "\",\"Left mouse botton: Close and save as " << G.name() << ".eps             Right mouse botton: Close\")\n";
	file2 << "tmp <- locator(1) \n";
	file2 << "graphics.off()\n";
	file2 << "if(length(tmp>0))\n";
	file2 << "{\n";
	file2 << "postscript(\"" << G.name() << ".eps\",width=" << width << ",height=" <<  height  << ",onefile=F,title=\"" << G.name() << "\",horizontal=T)\n";
	file2 << "image(1:dim(IM2)[1],1:dim(IM2)[2],as.matrix(IM2),ann=F,col=gray(100:10/110))\n";
	file2 << "title(\"" <<  G.name() << "\")\n";
	file2 << "graphics.off()\n";
	file2 << "}\n";
	file2 << "q()";
}


/* Parameter IO */

void StructSearch::IgnoreLines(istream &astream)
{
  char ch;

  while(astream >> ch)
  {
    if(ch=='#'){while(ch != '\n'){astream.get(ch);}}
    else 
      {
	astream.putback(ch);
	break;
      }
  }
}
  

void StructSearch::read_par(const char* filename)
{
	SScore.init(16,16);
	NullTP.init(4,4);
	StackMask.init(16,16);
	
	double tmp;
	long i,j;
	char ch;
        istringstream parfile(defaultpar(filename)); 
	
	IgnoreLines(parfile);                   //Ignores comments

	//Read SScore matrix and define StackMask	
	i=0;
	j=0;

	while(i < 16)
	{
	  while(j < 16) 
		{
		  if(parfile >> ch)
			{
			  if(ch == '#')
				{
				  parfile.putback(ch);
				  IgnoreLines(parfile);                   //Ignores comments
				}
			  else if(ch == '*')
			    {
			      StackMask(i,j) = 0;
			      j++;
			    }
			  else 			    
			    {
			      parfile.putback(ch);
			      if(parfile >> tmp)
				{
				  SScore(i,j) = tmp; 
				  StackMask(i,j) = 1;
				  j++;
				}
			      else
				{
				  error("Format problems with parameter file");
				}
			    }
			}
			  else
			    {
			      error("Format problems with parameter file");
			    }
			
		}
	  j=0;
	  i++;
	}	
	
        //Read loop penalty parameters. First Loop initiation, then loop extension, and then hairpin loop penalties
	
	IgnoreLines(parfile);                   //Ignores comments
	if(!(parfile >> Li)) error("Format problems with parameter file (loop)");
	IgnoreLines(parfile);                   //Ignores comments
	if(!(parfile >> Le)) error("Format problems with parameter file (loop)");
	
	IgnoreLines(parfile);                   //Ignores comments

	while(parfile >> ch)
	  {
	    parfile.putback(ch);
	    if(ch == '#') break;
	    else if(parfile >> tmp)
		  {
		     HPLoopPen.push_back(min(tmp, (double) 0));
		  }
	    else break;
	      
	  }

	if(HPLoopPen.size() < 2) {error("Format problems with parameter file (loop)");}
	
	//Hairpin loop penalty affine extrapolation
	
	if(hplooppen)
	  {
	    tmp = HPLoopPen.back() - HPLoopPen[HPLoopPen.size()-2];
	    while(HPLoopPen.size() < band+1)
	      {
		HPLoopPen.push_back(HPLoopPen.back() + tmp);
	      }
	  }
	else
	  {
	    HPLoopPen.assign(band, 0);
	  }


	//Read null model transition probabilities

	IgnoreLines(parfile);  

	for(i=0; i <4; i++)
	{
		j = 0;
		while((j < 4) && (parfile >> tmp)) 
		{
		  
		  NullTP(i,j) = max(tmp,(double) 0); 
		  j++;
		}
		if((j != 4)) error("Format problems with NullTP");
	}

	for(i=0;i<4;i++)
	  {
	    tmp = 0;
	    for(j=0;j<4;j++)
	      tmp += NullTP(i,j);
	    if(!(tmp == 1))
	      {
		if(warnings)
		  {
		    ofstream file("StemSearch.log", ios_base::app);
		    if (!file) error("Problems opening log-file");
		    time_t rawtime = time(NULL);
		    string logtime = ctime(&rawtime);
		    file << logtime.substr(0,24) << "\t\tNormalisation warning: Row " << i+1 << " in null model transition probability matrix does not sum to 1.\n";
		    file.close();
		  }
		for(j=0;j<4;j++)
		  NullTP(i,j) *= 1/tmp;
	      }
	  }


	//Read normalisation parameters. First with hairpin loop penalty and then without.
	
	IgnoreLines(parfile);                   //Ignores comments
	if(!(parfile >> lambda)) error("Format problems with parameter file (normalisation)");
	IgnoreLines(parfile);                   //Ignores comments
	if(!(parfile >> logK)) error("Format problems with parameter file (normalisation)");
	IgnoreLines(parfile);                   //Ignores comments
	if(!(parfile >> H)) error("Format problems with parameter file (normalisation)");


	if(!hplooppen)
	  {
	    IgnoreLines(parfile);                   //Ignores comments
	    if(!(parfile >> lambda)) error("Format problems with parameter file (normalisation)");
	    IgnoreLines(parfile);                   //Ignores comments
	    if(!(parfile >> logK)) error("Format problems with parameter file (normalisation)");
	    IgnoreLines(parfile);                   //Ignores comments
	    if(!(parfile >> H)) error("Format problems with parameter file (normalisation)");
	  }




}

string StructSearch::defaultpar(const char* filename)
{
  stringstream par;
  string temp_filename = filename;
  if(temp_filename == "")
    {
      par << "###################################################################################\n"
"# Parameter file, default.par, for StemSearch. Created Thu Sep 28 16:05:55 2006 by\n" 
"# the StemSearch R package 1.0. Hand-modified october 2007 (based on Human)               \n"
"###################################################################################\n"
"#\n"
"#Row and column order: aa ca ga ua ac cc gc uc ag cg gg ug au cu gu uu\n"
"#Stack and stack-entry transition scores:\n"
"*	*	*	0.8701	*	*	1.1399	*	*	3.0498	*	0.1305	0.441	*	-0.6776	*\n"
"*	*	*	1.1223	*	*	1.7685	*	*	3.2484	*	-0.4085	0.7871	*	-0.4057	*\n"
"*	*	*	2.0804	*	*	1.9847	*	*	2.344	*	0.1644	1.2058	*	*	*\n"
"*	*	*	1.5391	*	*	1.5054	*	*	3.231	*	0.0042	1.0368	*	0.2406	*\n"
"*	*	*	1.2469	*	*	1.9278	*	*	2.7633	*	*	1.3272	*	-0.8464	*\n"
"*	*	*	1.5309	*	*	1.378	*	*	2.7754	*	-0.5186	1.27	*	-0.2281	*\n"
"*	*	*	1.8378	*	*	1.8098	*	*	1.8195	*	-0.2658	1.5516	*	0.4393	*\n"
"*	*	*	1.3598	*	*	1.4456	*	*	2.7584	*	-1.2287	0.9113	*	-0.1273	*\n"
"*	*	*	-0.311	*	*	1.6933	*	*	2.9101	*	-0.4146	0.6268	*	-1.105	*\n"
"*	*	*	3.039	*	*	3.1212	*	*	4.5859	*	1.4517	2.6342	*	1.9283	*\n"
"*	*	*	1.4394	*	*	1.3494	*	*	2.2966	*	0.8589	1.2071	*	-0.0828	*\n"
"*	*	*	0.7321	*	*	1.2664	*	*	2.8432	*	-0.2346	0.7028	*	-1.3304	*\n"
"*	*	*	1.1065	*	*	1.354	*	*	3.1779	*	-0.1612	1.0155	*	0.0383	*\n"
"*	*	*	1.0994	*	*	1.6433	*	*	3.2444	*	0.1691	0.9386	*	0.0177	*\n"
"*	*	*	1.8263	*	*	1.8641	*	*	2.9624	*	0.5389	1.7055	*	0.276	*\n"
"*	*	*	0.5763	*	*	1.2465	*	*	3.1043	*	-0.2205	0.5051	*	-0.2822	*\n"
"\n"
"#Loop penalties\n"
"\n"
"-4\n"
"-2.1\n"
"0\n"
"0\n"
"0\n"
"-0.4\n"
"\n"
"#Null model\n"
"\n"
"0.3307 0.1708 0.2377 0.2607\n"
"0.3537 0.2558 0.0487 0.3417\n"
"0.2891 0.2094 0.256 0.2455\n"
"0.222 0.2011 0.2462 0.3307\n"
"\n"
"#Normalization\n"
"\n"
"0.3535\n"
"-0.3099\n"
"1\n"
"0.3672\n"
"-2.7048\n"
	"1\n";}
  else
    {
      ifstream parfile(filename);
      if(!parfile) error("Error when reading parameter file\n");
      char ch;
      while(parfile.get(ch)) par << ch;
    }

 

  return(par.str());
}

	
	
	
	
	
	
	
	
	
	
	
	
	

