/*  Copyright (C) 2008 Niels Richard Hansen. 

    This file is part of StemSearch.

    StemSearch is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    StemSearch is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MATRIX_INC_GUARD
#define MATRIX_INC_GUARD

#include<iostream>
#include<fstream>
#include<iomanip>
using namespace std;

template<class T> class Matrix 
{
 protected:
  int rows;
  int columns;
 public:
  virtual void resize(int, int) =0; 
  virtual void init(T = T());
  virtual void init(T*);
  virtual void init(int, int=1);
  virtual void init(T*, int, int);   
	
  virtual const T& operator()(int, int) const =0; 
  virtual T& operator()(int, int) =0;
	
  int nr_rows() const;
  int nr_columns() const;
		
  void error(char*) const;
	
  int operator==(const Matrix<T> &) const;
 
  void operator+=(const T &);
  void operator*=(const T &);
  void operator-=(const T &);

  void operator+=(const Matrix<T> &);
  void operator-=(const Matrix<T> &);
	
  void print(int =0, char ='r') const;
  void save(const char*,const char* =" ",const char* = "") const;

};


template<class T> class RectMatrix : public Matrix<T>
{
 protected:
  using Matrix<T>::rows;
  using Matrix<T>::columns;
  T** A;
 public:
  using Matrix<T>::error;
  RectMatrix(T = T());
  RectMatrix(int, int); 
  RectMatrix(T*,int,int);
  RectMatrix(const Matrix<T>&);
  RectMatrix(const RectMatrix<T>&);
  ~RectMatrix(); 
  RectMatrix<T> &operator=(const Matrix<T>&);
  RectMatrix<T> &operator=(const RectMatrix<T>&);
	
  void resize(int, int =1); 
	
  RectMatrix<T> operator+(const T &) const;
  RectMatrix<T> operator*(const T &) const;
  RectMatrix<T> operator-(const T &) const;

  RectMatrix<T> operator+(const RectMatrix<T> &) const; //range checked
  RectMatrix<T> operator*(const RectMatrix<T> &) const; //range checked
  RectMatrix<T> operator-(const RectMatrix<T> &) const; //range checked
	
  const T& operator()(int, int=0) const; 
  T& operator()(int, int=0);
  RectMatrix<T> trans() const;
};

template<class T> class UpBandMatrix : public Matrix<T>
{
 private:
  int width;
  using Matrix<T>::rows;
  using Matrix<T>::columns;
  T dummy1,dummy2;
  T** A;
 public:
  using Matrix<T>::error;
  UpBandMatrix(T = T());
  UpBandMatrix(int, int); 
  UpBandMatrix(int, int, int); 
  UpBandMatrix(T*,int,int);
  UpBandMatrix(T*,int,int,int);
  UpBandMatrix(const Matrix<T>&);
  UpBandMatrix(const UpBandMatrix<T>&);
  ~UpBandMatrix(); 
  UpBandMatrix<T> &operator=(const Matrix<T>&);
  UpBandMatrix<T> &operator=(const UpBandMatrix<T>&);

	
  void resize(int, int =1); 
  void resize(int, int, int);
  void init();
  void init(int,int);
  void init(int, int, int);
  void init(T*);
  void init(T*, int, int);
  void init(T*, int, int, int);
	
  UpBandMatrix<T> operator+(const T &) const;
  UpBandMatrix<T> operator*(const T &) const;
  UpBandMatrix<T> operator-(const T &) const;

  UpBandMatrix<T> operator+(const UpBandMatrix<T> &) const; //range checked
  UpBandMatrix<T> operator*(const UpBandMatrix<T> &) const; //range checked
  UpBandMatrix<T> operator-(const UpBandMatrix<T> &) const; //range checked
	
  const T& operator()(int, int=0) const; 
  T& operator()(int, int=0);
  UpBandMatrix<T> trans() const;
  void eff_save(const char*,const char* =" ",const char* = "") const;
};

template<class T> class LoBandMatrix : public UpBandMatrix<T>
{
  using UpBandMatrix<T>::width;
  using UpBandMatrix<T>::rows;
  using UpBandMatrix<T>::columns;
  T dummy1,dummy2;
  T** A;
 public:
  using Matrix<T>::error;
  const T& operator()(int, int=0) const; 
  T& operator()(int, int=0);
};

/*

IMPLEMENTATION OF Matrix TEMPLATE:
		 
*/

template<class T> void Matrix<T>::error(char error_text[]) const
{
  cout << "\n" << error_text;
  exit(1);
}

template<class T> int Matrix<T>::nr_rows() const
{
  return(rows);
}

template<class T> int Matrix<T>::nr_columns() const
{
  return(columns);
}

template<class T> void Matrix<T>::init(T t)
{
  int i,j;
  for (i=0; i<rows; i++)
    {
      for(j=0;j<columns;j++)
	operator()(i,j) = t;
    }
}
  
template<class T> void Matrix<T>::init(int r, int c)
{
  resize(r,c);
  init();
}
  
template<class T> void Matrix<T>::init(T* t)       //column major order
{
  int i,j;
  for (i=0; i<rows; i++)
    {
      for(j=0;j<columns;j++)
	operator()(i,j) = t[j+i*columns];
    }
}
  
template<class T> void Matrix<T>::init(T* t, int r, int c)       //column major order
{
  resize(r,c);
  init(t);
}
  

template<class T> int Matrix<T>::operator==(const Matrix<T> &B) const
{
  int i,j,eq=1;
  if ((rows != B.rows) & (columns != B.columns))
    error("Wrong dimensions in Matrix equal \n");               
  else 
    {
      for(i=0; (i<rows) && eq; i++)
	{
	  for(j=0; (j<columns) && eq; j++)
	    eq = (operator()(i,j) == B(i,j)); 
	}
    }
  return(eq);
}


template<class T> void Matrix<T>::operator+=(const T &x) 
{
  int i,j;
  for(i=0;i < rows; i++)
    {
      for(j=0; j<columns;j++)
	operator()(i,j) += x;
    }
}

template<class T> void Matrix<T>::operator*=(const T &x) 
{
  int i,j;
  for(i=0;i < rows; i++)
    {
      for(j=0; j<columns;j++)
	{
	  operator()(i,j) *= x;
	}
    }
}

template<class T> void Matrix<T>::operator-=(const T &b) 
{
  int i,j;
  for(i=0;i < rows; i++)
    {
      for(j=0; j<columns;j++)
	operator()(i,j) -= b;
    }
}

template<class T> void Matrix<T>::operator+=(const Matrix<T> &B) 
{
  int i,j;
  if ((rows != B.rows) & (columns != B.columns))
    error("Wrong dimensions in Matrix plus \n");                
  else 
    {
      for(i=0;i < rows; i++)
	{
	  for(j=0; j<columns;j++)
	    operator()(i,j) += B(i,j);
	}
    }
}



template<class T> void Matrix<T>::operator-=(const Matrix<T> &B) 
{
  int i,j;
  if ((rows != B.rows) & (columns != B.columns))
    error("Wrong dimensions in Matrix minus \n");               
  else 
    {
      for(i=0;i < rows; i++)
	{
	  for(j=0; j<columns;j++)
	    operator()(i,j) -= B(i,j);
	}
    }
}



template<class T>  void Matrix<T>::save(const char* filename,const char* sep,const char* end_of_line) const
{
  int i,j;
  ofstream file(filename);
	
  if (!file) error("cannot open file");
	
  file.setf(ios_base::right,ios_base::adjustfield);

  for(i=0;i<rows;i++)
    {
      for(j=0;j<columns-1;j++)
	{
	  file << setw(5) << operator()(i,j) << sep;
	}
      file << setw(5) << operator()(i,columns-1) << end_of_line << "\n";
    }
  file << "\n";
}

template<class T> void Matrix<T>::print(int w, char ch) const
{
  int i,j;
  switch (ch)
    {
    case 'l': 
      cout.setf(ios_base::left,ios_base::adjustfield);
      break;
    case 'r':
      cout.setf(ios_base::right,ios_base::adjustfield);
      break;
    }
  for(i=0;i< rows;i++)
    {
      for(j=0;j< columns;j++)
	cout << setw(w) << operator()(i,j) << " ";
      cout << "\n";
    }
  cout << "\n";
}

/*

IMPLEMENTATION OF RectMatrix TEMPLATE:
		 
*/

template<class T> RectMatrix<T>::RectMatrix(T t) 
{
  rows=1; columns=1;
  A = new T*[1];
  A[0] = new T[1];
  A[0][0] = t;
}

template<class T> RectMatrix<T>::RectMatrix(int r, int c) 
{
  int i;
  rows=r; columns=c;
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[columns];
}
	
template<class T> RectMatrix<T>::RectMatrix(T* t, int r, int c) //column major order
{
  int i,j;
  rows=r; columns=c;
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[columns];
  for (i=0; i<rows; i++)
    {
      for(j=0;j<columns;j++)
	A[i][j] = t[j+i*c];
    }
}
	
template<class T> RectMatrix<T>::RectMatrix(const RectMatrix<T> &B)
{
  int i,j;
  rows = B.rows;
  columns = B.columns;
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[columns];
  for (i=0;i<rows;i++)
    {
      for(j=0;j<columns;j++)
	A[i][j]=B(i,j);
    }
}
	
template<class T> RectMatrix<T>::RectMatrix(const Matrix<T> &B)
{
  int i,j;
  rows = B.rows;
  columns = B.columns;
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[columns];
  for (i=0;i<rows;i++)
    {
      for(j=0;j<columns;j++)
	A[i][j]=B(i,j);
    }
}
	
template<class T> RectMatrix<T>::~RectMatrix() 
{
  int i;
  for (i=0; i<rows; i++)
    delete [] A[i];
  delete [] A;
}

template<class T> void RectMatrix<T>::resize(int r, int c)                        //warning: This deletes the original Matrix
{ 
  int i;
  for (i=0; i<rows; i++)
    delete [] A[i];
  delete [] A;
  rows = r; columns = c;
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[columns];
}
		
template<class T> RectMatrix<T>& RectMatrix<T>::operator=(const RectMatrix<T> & B)
{
  if (this != &B)
    {
      int i,j;
      resize(B.rows,B.columns);
      for (i=0;i<rows;i++)
	{
	  for(j=0;j<columns;j++)
	    A[i][j] = B(i,j);
	}
    }
  return *this;
}
  
template<class T> RectMatrix<T>& RectMatrix<T>::operator=(const Matrix<T> & B)
{
  if (this != &B)
    {
      int i,j;
      resize(B.rows,B.columns);
      for (i=0;i<rows;i++)
	{
	  for(j=0;j<columns;j++)
	    A[i][j] = B(i,j);
	}
    }
  return *this;
}
  
template<class T> const T& RectMatrix<T>::operator()(int i, int j) const 
{
  return A[i][j];
} 

template<class T> T& RectMatrix<T>::operator()(int i, int j) 
{
  return A[i][j];
}


  
template<class T> RectMatrix<T> RectMatrix<T>::operator+(const T &x) const
{
  int i,j;
  RectMatrix<T> C(rows,columns);
  for(i=0;i < rows; i++)
    {
      for(j=0; j<columns;j++)
	C(i,j) = operator()(i,j) + x;
    }
  return(C);
}

template<class T> RectMatrix<T> RectMatrix<T>::operator*(const T &x) const
{
  int i,j;
  RectMatrix<T> C(rows,columns);
  for(i=0;i < rows; i++)
    {
      for(j=0; j<columns;j++)
	{
	  C(i,j) = x*operator()(i,j);
	}
    }
  return(C);
}

template<class T> RectMatrix<T> RectMatrix<T>::operator-(const T &x) const
{
  int i,j;
  RectMatrix<T> C(rows,columns);
  for(i=0;i < rows; i++)
    {
      for(j=0; j<columns;j++)
	C(i,j) = operator()(i,j) - x;
    }
  return(C);
}

template<class T> RectMatrix<T> RectMatrix<T>::operator+(const RectMatrix<T> &B) const
{
  int i,j;
  RectMatrix<T> C(rows,columns);
  if ((rows != B.rows) & (columns != B.columns))
    error("Wrong dimensions in RectMatrix plus \n");                
  else 
    {
      for(i=0;i < rows; i++)
	{
	  for(j=0; j<columns;j++)
	    C(i,j) = operator()(i,j) + B(i,j);
	}
    }
  return(C);
}


template<class T> RectMatrix<T> RectMatrix<T>::operator*(const RectMatrix<T> &B) const
{
  int i,j,k;
  RectMatrix<T> C(rows,B.columns);
  if (columns != B.rows) 
    error("Wrong dimensions in RectMatrix multiplication \n");
  else 
    { 
      C.init();
      for(i=0;i < C.rows; i++)
	{
	  for(j=0; j<C.columns;j++)
	    {
	      for(k=0;k<columns;k++)
		C(i,j) += operator()(i,k)*B(k,j);
	    }
	}
    }
  return(C);
}

template<class T> RectMatrix<T> RectMatrix<T>::operator-(const RectMatrix<T> &B) const
{
  int i,j;
  RectMatrix<T> C(rows,columns);
  if ((rows != B.rows) & (columns != B.columns))
    error("Wrong dimensions in RectMatrix minus \n");               
  else 
    {
      for(i=0;i < rows; i++)
	{
	  for(j=0; j<columns;j++)
	    C(i,j) = operator()(i,j) - B(i,j);
	}
    }
  return(C);
}
  
template<class T> RectMatrix<T> RectMatrix<T>::trans() const
{             
  int i,j;
  RectMatrix<T> B(columns,rows);
  for (i=0;i<columns;i++)
    {
      for(j=0;j<rows;j++)
	B(i,j) = operator()(j,i);
    }
  return(B);
}

/*

IMPLEMENTATION OF UpBandMatrix TEMPLATE:
		 
*/

template<class T> UpBandMatrix<T>::UpBandMatrix(T t) 
{
  rows=1; columns=1; width=1;
  A = new T*[1];
  A[0] = new T[1];
  A[0][0] = t;
  dummy1 = T();
  dummy2 = T();
}

template<class T> UpBandMatrix<T>::UpBandMatrix(int r, int c) 
{
  UpBandMatrix(r,c,c);
}
	
template<class T> UpBandMatrix<T>::UpBandMatrix(int r, int c, int w) 
{
  int i;
  rows=r; columns=c; width=w;
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[width];
  dummy1 = T();
  dummy2 = T();
}
	
template<class T> UpBandMatrix<T>::UpBandMatrix(T* t, int r, int c) //column major order
{
  UpBandMatrix(t,r,c,c);
}
	
template<class T> UpBandMatrix<T>::UpBandMatrix(T* t, int r, int c, int w) //column major order
{
  int i,j;
  rows=r; columns=c; width=w;
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[width];
  for (i=0; i<rows; i++)
    {
      for(j=0;(j<width) && (j + i < columns);j++)
	A[i][j] = t[j+i*width];
    }
  dummy1 = T();
  dummy2 = T();
}
	
template<class T> UpBandMatrix<T>::UpBandMatrix(const Matrix<T> &B)
{
  int i,j;
  rows = B.nr_rows();
  columns = B.nr_columns();
  width = B.nr_columns();
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[width];
  for (i=0;i<rows;i++)
    {
      for(j=0;(j<width) && (j + i < columns);j++)
	A[i][j]=B(i,j + i);
    }
}

template<class T> UpBandMatrix<T>::UpBandMatrix(const UpBandMatrix<T> &B)
{
  int i,j;
  rows = B.rows;
  columns = B.columns;
  width = B.width;
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[width];
  for (i=0;i<rows;i++)
    {
      for(j=0;(j<width) && (j + i < columns);j++)
	A[i][j]=B.A[i][j];
    }
}
	
template<class T> UpBandMatrix<T>::~UpBandMatrix() 
{
  int i;
  for (i=0; i<rows; i++)
    delete [] A[i];
  delete [] A;
}
	
template<class T> UpBandMatrix<T>& UpBandMatrix<T>::operator=(const Matrix<T> &B)
{
  if (this != &B)
    {
      int i,j;
      for (i=0; i<rows; i++)
	delete [] A[i];
      delete [] A;
      rows = B.nr_rows();
      columns = B.nr_columns();
      width = B.nr_columns();
      A = new T*[rows];
      for (i=0;i<rows;i++)
	A[i] = new T[width];
      for (i=0;i<rows;i++)
	{
	  for(j=0;(j<width) && (j + i < columns);j++)
	    A[i][j] = B(i,j+i);
	}
    }
  return *this;
}
  
template<class T> UpBandMatrix<T>& UpBandMatrix<T>::operator=(const UpBandMatrix<T> &B)
{
  if (this != &B)
    {
      int i,j;
      for (i=0; i<rows; i++)
	delete [] A[i];
      delete [] A;
      rows = B.rows;
      columns = B.columns;
      width = B.width;
      A = new T*[rows];
      for (i=0;i<rows;i++)
	A[i] = new T[width];
      for (i=0;i<rows;i++)
	{
	  for(j=0;(j<width) && (j + i < columns);j++)
	    A[i][j] = B.A[i][j];
	}
    }
  return *this;
}
  
template<class T> const T& UpBandMatrix<T>::operator()(int i, int j) const 
{
  if ((0 <=i) && (j < columns) && (i <= j) && (j-i < width))
    return A[i][j-i];
  else
    return dummy2;
} 

template<class T> T& UpBandMatrix<T>::operator()(int i, int j) 
{
  if ((j-i >=0) && (j-i < width))
    return A[i][j-i];
  else
    return dummy1;
}

template<class T> void UpBandMatrix<T>::resize(int r, int c, int w)                        //warning: This deletes the original Matrix
{ 
  int i;
  for (i=0; i<rows; i++)
    delete [] A[i];
  delete [] A;
  rows = r; columns = c; width=w;
  A = new T*[rows];
  for (i=0;i<rows;i++)
    A[i] = new T[width];
}
  
template<class T> void UpBandMatrix<T>::resize(int r, int c)                        //warning: This deletes the original Matrix
{ 
  resize(r,c,c);
}

template<class T> void UpBandMatrix<T>::init()
{
  Matrix<T>::init();
}

template<class T> void UpBandMatrix<T>::init(int r, int c, int w)
{
  resize(r,c,w);
  init();
}

template<class T> void UpBandMatrix<T>::init(int r, int c)
{
  resize(r,c,c);
  init();
}

template<class T> void UpBandMatrix<T>::init(T* t)       //column major order
{
  int i,j;
  for (i=0; i<rows; i++)
    {
      for(j=0;j<width;j++)
	A[i][j] = t[j+i*width];
    }
}
 
template<class T> void UpBandMatrix<T>::init(T* t, int r, int c)       //column major order
{
  resize(r,c,c);
  init(t);
}
  
template<class T> void UpBandMatrix<T>::init(T* t, int r, int c, int w)       //column major order
{
  resize(r,c,w);
  init(t);
}


template<class T> UpBandMatrix<T> UpBandMatrix<T>::operator+(const T &x) const
{
  int i,j;
  UpBandMatrix<T> C(rows,columns);
  for(i=0;i < rows; i++)
    {
      for(j=0; j<columns;j++)
	C(i,j) = operator()(i,j) + x;
    }
  return(C);
}

template<class T> UpBandMatrix<T> UpBandMatrix<T>::operator*(const T &x) const
{
  int i,j;
  UpBandMatrix<T> C(rows,columns);
  for(i=0;i < rows; i++)
    {
      for(j=0; j<columns;j++)
	{
	  C(i,j) = x*operator()(i,j);
	}
    }
  return(C);
}

template<class T> UpBandMatrix<T> UpBandMatrix<T>::operator-(const T &x) const
{
  int i,j;
  UpBandMatrix<T> C(rows,columns);
  for(i=0;i < rows; i++)
    {
      for(j=0; j<columns;j++)
	C(i,j) = operator()(i,j) - x;
    }
  return(C);
}

template<class T> UpBandMatrix<T> UpBandMatrix<T>::operator+(const UpBandMatrix<T> &B) const
{
  int i,j;
  UpBandMatrix<T> C(rows,columns);
  if ((rows != B.rows) & (columns != B.columns))
    error("Wrong dimensions in UpBandMatrix plus \n");                
  else 
    {
      for(i=0;i < rows; i++)
	{
	  for(j=0; j<columns;j++)
	    C(i,j) = operator()(i,j) + B(i,j);
	}
    }
  return(C);
}


template<class T> UpBandMatrix<T> UpBandMatrix<T>::operator*(const UpBandMatrix<T> &B) const
{
  int i,j,k;
  UpBandMatrix<T> C(rows,B.columns);
  if (columns != B.rows) 
    error("Wrong dimensions in UpBandMatrix multiplication \n");
  else 
    { 
      C.init();
      for(i=0;i < C.rows; i++)
	{
	  for(j=0; j<C.columns;j++)
	    {
	      for(k=0;k<columns;k++)
		C(i,j) += operator()(i,k)*B(k,j);
	    }
	}
    }
  return(C);
}

template<class T> UpBandMatrix<T> UpBandMatrix<T>::operator-(const UpBandMatrix<T> &B) const
{
  int i,j;
  UpBandMatrix<T> C(rows,columns);
  if ((rows != B.rows) & (columns != B.columns))
    error("Wrong dimensions in UpBandMatrix minus \n");               
  else 
    {
      for(i=0;i < rows; i++)
	{
	  for(j=0; j<columns;j++)
	    C(i,j) = operator()(i,j) - B(i,j);
	}
    }
  return(C);
}
  
template<class T> UpBandMatrix<T> UpBandMatrix<T>::trans() const
{             
  int i,j;
  UpBandMatrix<T> B(columns,rows);
  for (i=0;i<columns;i++)
    {
      for(j=0;j<rows;j++)
	B(i,j) = operator()(j,i);
    }
  return(B);
}

template<class T>  void UpBandMatrix<T>::eff_save(const char* filename,const char* sep,const char* end_of_line) const
{
  int i,j;
  ofstream file(filename);
	
  if (!file) error("cannot open file");
	
  file.setf(ios_base::right,ios_base::adjustfield);
	
  int tmp = 2*((width+1)/2) - 2; 		//=width-2 if width is even and width-1 if width is odd
	
  for(j=tmp;j>=0;j=j-2)
    {
      for(i=0;i<j;i++)
	file << setw(5) << 0 << sep;
      for(i=0;i<rows-j;i++)
	{
	  file << setw(5) << operator()(i,i+j) << sep;
	  file << setw(5) << operator()(i,i+j+1) << sep;
	}
      for(i=0;i<j;i++)
	file << setw(5) << 0 << sep;
      file << end_of_line << "\n";
    }
  file << "\n";
}


/*

IMPLEMENTATION OF LoBandMatrix TEMPLATE:
		 
*/

template<class T> const T& LoBandMatrix<T>::operator()(int i, int j) const 
{
  return UpBandMatrix<T>::operator()(j,i);
} 

template<class T> T& LoBandMatrix<T>::operator()(int i, int j) 
{
  return UpBandMatrix<T>::operator()(j,i);
}

#endif  //MATRIX_INC_GUARD
